﻿//    License: Microsoft Public License (Ms-PL) 

using System;
using System.IO;
using System.Linq;
using System.Diagnostics;
using Lucene.Net.Store;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.StorageClient;
using Directory = Lucene.Net.Store.Directory;

namespace Lucene.Linq.Azure
{
    public class AzureDirectory : Directory
    {
        private readonly string _catalog;
        private CloudBlobClient _blobClient;
        private Directory _cacheDirectory;

        #region CTOR
        public AzureDirectory(CloudStorageAccount storageAccount) :
            this(storageAccount, null, null)
        { }

        /// <summary>
        /// Create AzureDirectory
        /// </summary>
        /// <param name="storageAccount">staorage account to use</param>
        /// <param name="catalog">name of catalog (folder in blob storage)</param>
        /// <remarks>Default local cache is to use file system in user/appdata/AzureDirectory/Catalog</remarks>
        public AzureDirectory(
            CloudStorageAccount storageAccount,
            string catalog)
            : this(storageAccount, catalog, null)
        { }

        /// <summary>
        /// Create an AzureDirectory
        /// </summary>
        /// <param name="storageAccount">storage account to use</param>
        /// <param name="catalog">name of catalog (folder in blob storage)</param>
        /// <param name="cacheDirectory">local Directory object to use for local cache</param>
        public AzureDirectory(
            CloudStorageAccount storageAccount,
            string catalog,
            Directory cacheDirectory)
        {
            if (storageAccount == null)
            {
                Trace.TraceError("storageAccount argument cannot be null");
                throw new ArgumentNullException("storageAccount");
            }

            _catalog = string.IsNullOrEmpty(catalog) ? "lucene" : catalog.ToLower();

            _blobClient = storageAccount.CreateCloudBlobClient();
            InitCacheDirectory(cacheDirectory);
        }

        public CloudBlobContainer BlobContainer { get; private set; }

#if COMPRESSBLOBS
        public bool CompressBlobs
        {
            get;
            set;
        }
#endif
        public void ClearCache()
        {
            Trace.TraceInformation("Clearing cache");
            foreach (string file in _cacheDirectory.ListAll())
            {
                _cacheDirectory.DeleteFile(file);
            }
        }

        public Directory CacheDirectory
        {
            get
            {
                return _cacheDirectory;
            }
            set
            {
                _cacheDirectory = value;
            }
        }
        #endregion

        #region internal methods
        private void InitCacheDirectory(Directory cacheDirectory)
        {
#if COMPRESSBLOBS
            CompressBlobs = true;
#endif
            if (cacheDirectory != null)
            {
                // save it off
                _cacheDirectory = cacheDirectory;
            }
            else
            {
                string cachePath = Path.Combine(Environment.ExpandEnvironmentVariables("%temp%"), "AzureDirectory");
                DirectoryInfo azureDir = new DirectoryInfo(cachePath);
                if (!azureDir.Exists)
                    azureDir.Create();

                string catalogPath = Path.Combine(cachePath, _catalog);
                DirectoryInfo catalogDir = new DirectoryInfo(catalogPath);
                if (!catalogDir.Exists)
                    catalogDir.Create();

                _cacheDirectory = FSDirectory.Open(catalogDir);
            }

            CreateContainer();
        }

        public void CreateContainer()
        {
            BlobContainer = _blobClient.GetContainerReference(_catalog);

            // create it if it does not exist
            BlobContainer.CreateIfNotExist();
        }
        #endregion

        #region DIRECTORY METHODS
        /// <summary>Returns an array of strings, one for each file in the directory. </summary>
        [Obsolete("For some Directory implementations (FSDirectory}, and its subclasses), this method silently filters its results to include only index files.  Please use ListAll instead, which does no filtering. ")]
        public override String[] List()
        {
            var results = from blob in BlobContainer.ListBlobs()
                          select blob.Uri.AbsolutePath.Substring(blob.Uri.AbsolutePath.LastIndexOf('/') + 1);
            return results.ToArray();
        }

        /// <summary>Returns true if a file with the given name exists. </summary>
        public override bool FileExists(String name)
        {
            // this always comes from the server
            try
            {
                var blob = BlobContainer.GetBlobReference(name);
                blob.FetchAttributes();
                return true;
            }
            catch
            {
                return false;
            }
        }

        /// <summary>Returns the time the named file was last modified. </summary>
        public override long FileModified(String name)
        {
            // this always has to come from the server
            try
            {
                var blob = BlobContainer.GetBlobReference(name);
                blob.FetchAttributes();
                return blob.Properties.LastModifiedUtc.ToFileTimeUtc();
            }
            catch
            {
                return 0L;
            }
        }

        /// <summary>Set the modified time of an existing file to now. </summary>
        public override void TouchFile(String name)
        {
            //BlobProperties props = _blobContainer.GetBlobProperties(name);
            //_blobContainer.UpdateBlobMetadata(props);
            // I have no idea what the semantics of this should be...hmmmm...
            // we never seem to get called
            _cacheDirectory.TouchFile(name);
            //SetCachedBlobProperties(props);
        }

        /// <summary>Removes an existing file in the directory. </summary>
        public override void DeleteFile(String name)
        {
            var blob = BlobContainer.GetBlobReference(name);
            blob.DeleteIfExists();
            Trace.TraceInformation(String.Format("DELETE {0}/{1}", BlobContainer.Uri, name));

            if (_cacheDirectory.FileExists(name + ".blob"))
                _cacheDirectory.DeleteFile(name + ".blob");

            if (_cacheDirectory.FileExists(name))
                _cacheDirectory.DeleteFile(name);
        }

        /// <summary>Renames an existing file in the directory.
        /// If a file already exists with the new name, then it is replaced.
        /// This replacement should be atomic. 
        /// </summary>
        [Obsolete]
        public override void RenameFile(String from, String to)
        {
            try
            {
                var blobFrom = BlobContainer.GetBlobReference(from);
                var blobTo = BlobContainer.GetBlobReference(to);
                blobTo.CopyFromBlob(blobFrom);
                blobFrom.DeleteIfExists();

                // we delete and force a redownload, since we can't do this in an atomic way
                if (_cacheDirectory.FileExists(from))
                    _cacheDirectory.RenameFile(from, to);

                // drop old cached data as it's wrong now
                if (_cacheDirectory.FileExists(from + ".blob"))
                    _cacheDirectory.DeleteFile(from + ".blob");
            }
            catch (Exception ex)
            {
                Trace.TraceError("Rename file {0} to {1} failed with exception {2}", from, to, ex.ToString());
            }
        }

        /// <summary>Returns the length of a file in the directory. </summary>
        public override long FileLength(String name)
        {
            var blob = BlobContainer.GetBlobReference(name);
            blob.FetchAttributes();

            // index files may be compressed so the actual length is stored in metatdata
            long blobLength;
            return long.TryParse(blob.Metadata["CachedLength"], out blobLength)
                ? blobLength
                : blob.Properties.Length;
        }

        /// <summary>Creates a new, empty file in the directory with the given name.
        /// Returns a stream writing this file. 
        /// </summary>
        public override IndexOutput CreateOutput(String name)
        {
            var blob = BlobContainer.GetBlobReference(name);
            return new AzureIndexOutput(this, blob);
        }

        /// <summary>Returns a stream reading an existing file.</summary>
        public override IndexInput OpenInput(String name)
        {
            try
            {
                CloudBlob blob = BlobContainer.GetBlobReference(name);
                blob.FetchAttributes();
                AzureIndexInput input = new AzureIndexInput(this, blob);
                return input;
            }
            catch (Exception err)
            {
                var ex = new FileNotFoundException(name, err);
                Trace.TraceError(ex.ToString());
                throw ex;
            }
        }

        /// <summary>Construct a {@link Lock}.</summary>
        /// <param name="name">the name of the lock file
        /// </param>
        public override Lock MakeLock(String name)
        {
            _lock = name;
            return new AzureLock(name, this);
        }

        public override void ClearLock(string name)
        {
            var blob = BlobContainer.GetBlobReference(name);
            blob.DeleteIfExists();
            _cacheDirectory.ClearLock(name);
            _lock = string.Empty;
        }

        private string _lock = string.Empty;

        /// <summary>Closes the store. </summary>
        public override void Close()
        {
            if (!String.IsNullOrEmpty(_lock))
                ClearLock(_lock);
            BlobContainer = null;
            _blobClient = null;
        }
        #endregion

        #region Azure specific methods
#if COMPRESSBLOBS
        public virtual bool ShouldCompressFile(string path)
        {
            if (!CompressBlobs)
                return false;

            string ext = Path.GetExtension(path);
            switch (ext)
            {
                case ".cfs":
                case ".fdt":
                case ".fdx":
                case ".frq":
                case ".tis":
                case ".tii":
                case ".nrm":
                case ".tvx":
                case ".tvd":
                case ".tvf":
                case ".prx":
                    return true;
                default:
                    return false;
            }
        }
#endif
        public StreamInput OpenCachedInputAsStream(string name)
        {
            return new StreamInput(CacheDirectory.OpenInput(name));
        }

        public StreamOutput CreateCachedOutputAsStream(string name)
        {
            return new StreamOutput(CacheDirectory.CreateOutput(name));
        }

        #endregion

        /// <summary>
        /// Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources.
        /// </summary>
        /// <filterpriority>2</filterpriority>
        public override void Dispose()
        {
            Close();
        }
    }

}
