﻿//    License: Microsoft Public License (Ms-PL) 

using Lucene.Net.Store;
using Microsoft.WindowsAzure.StorageClient;
using System.Diagnostics;

namespace Lucene.Linq.Azure
{
    /// <summary>
    /// Implements lock semantics on AzureDirectory via a blob
    /// </summary>
    public class AzureLock : Lock
    {
        private readonly string _lockFile;
        private readonly AzureDirectory _azureDirectory;

        public AzureLock(string lockFile, AzureDirectory directory)
        {
            _lockFile = lockFile;
            _azureDirectory = directory;
        }

        #region Lock methods
        override public bool IsLocked()
        {
            var blob = _azureDirectory.BlobContainer.GetBlobReference(_lockFile);
            try
            {
                blob.FetchAttributes();
                return true;
            }
            catch (StorageClientException err)
            {
                if (err.ErrorCode != StorageErrorCode.ResourceNotFound)
                    Trace.TraceError(err.ToString());
                return false;
            }
        }

        public override bool Obtain()
        {
            try
            {
                if (IsLocked())
                    return false;

                var blob = _azureDirectory.BlobContainer.GetBlobReference(_lockFile);
                blob.UploadText("lock");
                Trace.TraceInformation(string.Format("Obtained lock {0}", _lockFile));
                return true;
            }
            catch (StorageClientException err)
            {
                Trace.TraceError(err.ToString());
                switch (err.ErrorCode)
                {
                    case StorageErrorCode.ContainerNotFound:
                        // container is missing, we should create it.
                        _azureDirectory.BlobContainer.Delete();
                        _azureDirectory.CreateContainer();
                        return Obtain();
                }
                return false;
            }
        }

        public override void Release()
        {
            Trace.TraceInformation(string.Format("Releasing lock {0}", _lockFile));
            var blob = _azureDirectory.BlobContainer.GetBlobReference(_lockFile);
            blob.DeleteIfExists();
        }
        #endregion

        public override System.String ToString()
        {
            return "Lock@" + _lockFile;
        }
    }
}
