﻿using System;
using System.IO;
using System.Linq;
using Lucene.Linq.VSTests.Data.CodeFirst;
using Lucene.Linq.VSTests.Helpers;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Lucene.Linq.Storage.EntityFramework;
using Customer = Lucene.Linq.VSTests.DatabaseIntegrations.CodeFirst.Customer;
using NorthwindDbContext = Lucene.Linq.VSTests.DatabaseIntegrations.CodeFirst.NorthwindDbContext;
using Order = Lucene.Linq.VSTests.DatabaseIntegrations.CodeFirst.Order;

namespace Lucene.Linq.VSTests
{
    [TestClass]
    public class TestEfDatabaseIndexSet
    {
        #region Generators

        private static readonly Random random = new Random();

        private string RandomPath()
        {
            return Path.Combine(Environment.CurrentDirectory, "index_" + random.String(30));
        }

        #endregion

        #region Ctors

        [TestMethod, ExpectedException(typeof (ArgumentNullException))]
        public void Construct_should_throw_given_null_data_context()
        {
            // db set should instantiate the db instead

            new EfDatabaseIndexSet<NorthwindDbContext>(null);

            new EfDatabaseIndexSet<NorthwindDbContext>(RandomPath(), null);

            new EfDatabaseIndexSet<NorthwindDbContext>(new DirectoryInfo(RandomPath()), null);
        }

        [TestMethod]
        public void Construct_should_add_indexes_to_set()
        {
            // db set should instantiate the db instead
            var dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new NorthwindDbContext());
            Assert.IsNotNull(dbi);
            Assert.IsNotNull(dbi.DataContext);
            Assert.IsTrue(dbi.Count == 2);

            dbi = new EfDatabaseIndexSet<NorthwindDbContext>(RandomPath(), new NorthwindDbContext());
            Assert.IsNotNull(dbi);
            Assert.IsNotNull(dbi.DataContext);
            Assert.IsTrue(dbi.Count == 2);

            dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new DirectoryInfo(RandomPath()), new NorthwindDbContext());
            Assert.IsNotNull(dbi);
            Assert.IsNotNull(dbi.DataContext);
            Assert.IsTrue(dbi.Count == 2);
        }

        [TestMethod]
        public void Construct_on_data_context_with_missing_document_attributes_should_add_only_relevant_indexes_to_set()
        {
            // db set should instantiate the db instead
            var dbi = new EfDatabaseIndexSet<Data.CodeFirst.NorthwindDbContext>(new Data.CodeFirst.NorthwindDbContext());
            Assert.IsNotNull(dbi.DataContext);
            Assert.IsTrue(dbi.Count == 1);
            Assert.IsTrue(dbi[typeof (Data.CodeFirst.Customer)].Name == "Customer");
        }

        [TestMethod, ExpectedException(typeof (ArgumentNullException))]
        public void Construct_throws_given_null_path_or_directory()
        {
            var dbi1 = new EfDatabaseIndexSet<NorthwindDbContext>("",
                                                                  new NorthwindDbContext());


            var dbi2 = new EfDatabaseIndexSet<NorthwindDbContext>((string)null,
                                                                  new NorthwindDbContext());


            var dbi3 = new EfDatabaseIndexSet<NorthwindDbContext>((DirectoryInfo)null,
                                                                  null);
        }

        #endregion

        #region Write

        [TestMethod]
        [Description("This indexes the Customer and Orders tables of the Northwind database and will take a while")]
        public void Write_all_should_add_to_index()
        {
            var dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new NorthwindDbContext());

            //ensure our db isn't empty otherwise we would get false positivies
            Assert.IsTrue(dbi.DataContext.Customers.Any());

            dbi.Write();

            // check that all rows were added as documents to each index
            Assert.IsTrue(
                dbi.Get<Customer>().Count == dbi.DataContext.Customers.Count()
                );

            Assert.IsTrue(
                dbi.Get<Order>().Count == dbi.DataContext.Orders.Count()
                );
        }

        [TestMethod]
        [Description("This indexes the Customer table of the Northwind database")]
        public void Write_single_table_should_add_to_index()
        {
            
            var dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new NorthwindDbContext());
            
            //ensure our db isn't empty otherwise we would get false positivies
            Assert.IsTrue(dbi.DataContext.Customers.Any());

            dbi.Write<Customer>();

            // check that all rows were added as documents to each index
            Assert.IsTrue(
                dbi.Get<Customer>().Count == dbi.DataContext.Customers.Count()
                );

            
        }

        [TestMethod]
        [Description("This attempts to write an index for a generic table type that doesn't exist within the dbcontext")]
        [ExpectedException(typeof(ArgumentException))]
        public void Write_single_table_using_generic_that_doesnt_exist_throws_arguement_exception()
        {

            var dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new NorthwindDbContext());

            dbi.Write<FakeTable>();
        }

        [TestMethod]
        [Description("This attempts to write an index for a nongeneric table type that doesn't exist within the dbcontext")]
        [ExpectedException(typeof(ArgumentException))]
        public void Write_single_table_using_nongeneric_that_doesnt_exist_throws_arguement_exception()
        {

            var dbi = new EfDatabaseIndexSet<NorthwindDbContext>(new NorthwindDbContext());

            dbi.Write(typeof(FakeTable));
        }

        #endregion
    }
}