using System;
using System.Collections;
using System.IO;
using System.Reflection;
using System.Xml.Linq;

namespace Lucene.Linq.Utility
{
    /// <summary>
    /// 
    /// </summary>
    public class ObjectDumper
    {
        private const ConsoleColor CIdent = ConsoleColor.Cyan;
        private const ConsoleColor CValue = ConsoleColor.Magenta;
        private readonly int _depth;

        private readonly TextWriter _writer;
        private int _level;
        private int _pos;

        private ObjectDumper(int depth)
        {
            _writer = Console.Out;
            _depth = depth;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="o"></param>
        public static void Write(object o)
        {
            Write(o, 0);
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="o"></param>
        /// <param name="depth"></param>
        public static void Write(object o, int depth)
        {
            var dumper = new ObjectDumper(depth);
            dumper.WriteObject(null, o);
        }

        private void Write(string s)
        {
            if (s != null)
            {
                _writer.Write(s);
                _pos += s.Length;
            }
        }

        private void Write(string s, ConsoleColor c)
        {
            ConsoleColor temp = Console.ForegroundColor;
            Console.ForegroundColor = c;
            Write(s);
            Console.ForegroundColor = temp;
        }

        private void WriteIndent()
        {
            for (int i = 0; i < _level; i++) _writer.Write("  ");
        }

        private void WriteLine()
        {
            _writer.WriteLine();
            _pos = 0;
        }

        private void WriteTab()
        {
            Write("  ");
            while (_pos % 8 != 0) Write(" ");
        }

        private void WriteObject(string prefix, object o)
        {
            if (o == null || o is ValueType || o is string || o is XElement)
            {
                WriteIndent();
                Write(prefix);
                WriteValue(o);
                WriteLine();
            }
            else if (o is IEnumerable)
            {
                foreach (object element in (IEnumerable)o)
                {
                    if (element is IEnumerable && !(element is string))
                    {
                        WriteIndent();
                        Write(prefix);
                        Write("...");
                        WriteLine();
                        if (_level < _depth)
                        {
                            _level++;
                            WriteObject(prefix, element);
                            _level--;
                        }
                    }
                    else
                    {
                        WriteObject(prefix, element);
                    }
                }
            }
            else
            {
                MemberInfo[] members = o.GetType().GetMembers(BindingFlags.Public | BindingFlags.Instance);
                WriteIndent();
                Write(prefix);
                bool propWritten = false;
                foreach (MemberInfo m in members)
                {
                    var f = m as FieldInfo;
                    var p = m as PropertyInfo;
                    if (f != null || p != null)
                    {
                        if (propWritten)
                        {
                            WriteTab();
                        }
                        else
                        {
                            propWritten = true;
                        }
                        Write(m.Name, CIdent);
                        Write("=");
                        Type t = f != null ? f.FieldType : p.PropertyType;
                        if (t.IsValueType || t == typeof(string))
                        {
                            WriteValue(f != null ? f.GetValue(o) : p.GetValue(o, null));
                        }
                        else
                        {
                            Write(typeof(IEnumerable).IsAssignableFrom(t) ? "..." : "{ }");
                        }
                    }
                }
                if (propWritten) WriteLine();
                if (_level < _depth)
                {
                    foreach (MemberInfo m in members)
                    {
                        var f = m as FieldInfo;
                        var p = m as PropertyInfo;
                        if (f != null || p != null)
                        {
                            Type t = f != null ? f.FieldType : p.PropertyType;
                            if (!(t.IsValueType || t == typeof(string)))
                            {
                                object value = f != null ? f.GetValue(o) : p.GetValue(o, null);
                                if (value != null)
                                {
                                    _level++;
                                    WriteObject(m.Name + ": ", value);
                                    _level--;
                                }
                            }
                        }
                    }
                }
            }
        }

        private void WriteValue(object o)
        {
            if (o == null)
            {
                Write("null");
            }
            else if (o is DateTime)
            {
                Write(((DateTime)o).ToShortDateString(), CValue);
            }
            else if (o is ValueType || o is string || o is XElement)
            {
                Write(o.ToString(), CValue);
            }
            else if (o is IEnumerable)
            {
                Write("...");
            }
            else
            {
                Write("{ }");
            }
        }
    }
}